@@ -1,23 +1,21 @@
 /*
- * This file is part of oFono - Open Source Telephony
  *
- * Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
+ *  oFono - Open Source Telephony
  *
- * Contact: Aki Niemi <aki.niemi@nokia.com>
+ *  Copyright (C) 2009-2010 Nokia Corporation and/or its subsidiary(-ies).
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * version 2 as published by the Free Software Foundation.
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License version 2 as
+ *  published by the Free Software Foundation.
  *
- * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
- * 02110-1301 USA
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  *
  */
 
@@ -50,6 +48,7 @@ struct netreg_data {
 	guint8 last_reg_mode;
 	guint8 rat;
 	guint8 gsm_compact;
+	guint8 strength;
 };
 
 static inline guint8 *mccmnc_to_bcd(const char *mcc, const char *mnc,
@@ -181,8 +180,9 @@ static gboolean decode_reg_status(struct netreg_data *nd, const guint8 *msg,
 	return TRUE;
 }
 
-static void reg_status_ind_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static void reg_status_ind_cb(GIsiClient *client,
+				const void *restrict data, size_t len,
+				uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_netreg *netreg = opaque;
@@ -199,11 +199,19 @@ static void reg_status_ind_cb(GIsiClient *client, const void *restrict data,
 	if (decode_reg_status(nd, msg+3, len-3, &status, &lac, &ci, &tech)) {
 		status = isi_status_to_at_status(status);
 		ofono_netreg_status_notify(netreg, status, lac, ci, tech);
+
+		/*
+		 * Make sure the core is also informed of current
+		 * signal strength, as it can be received before
+		 * registration status.
+		 */
+		ofono_netreg_strength_notify(netreg, nd->strength);
 	}
 }
 
-static bool reg_status_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean reg_status_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -222,7 +230,7 @@ static bool reg_status_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (len < 3 || msg[0] != NET_REG_STATUS_GET_RESP)
-		goto error;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -241,7 +249,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_registration_status(struct ofono_netreg *netreg,
@@ -268,8 +276,9 @@ error:
 	g_free(cbd);
 }
 
-static bool name_get_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean name_get_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -288,7 +297,7 @@ static bool name_get_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (len < 3 || msg[0] != NET_OPER_NAME_READ_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -302,7 +311,8 @@ static bool name_get_resp_cb(GIsiClient *client, const void *restrict data,
 		switch (g_isi_sb_iter_get_id(&iter)) {
 
 		case NET_GSM_OPERATOR_INFO:
-			if (!g_isi_sb_iter_get_oper_code(&iter, op.mcc, op.mnc, 2))
+			if (!g_isi_sb_iter_get_oper_code(&iter, op.mcc,
+								op.mnc, 2))
 				goto error;
 			break;
 
@@ -338,7 +348,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 
@@ -372,8 +382,9 @@ error:
 }
 
 
-static bool available_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean available_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -391,7 +402,7 @@ static bool available_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (len < 3 || msg[0] != NET_AVAILABLE_GET_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -402,25 +413,23 @@ static bool available_resp_cb(GIsiClient *client, const void *restrict data,
 	total = msg[2] / 2;
 	list = alloca(total * sizeof(struct ofono_network_operator));
 
-	g_isi_sb_iter_init(&iter, msg, len, 3);
+	for (g_isi_sb_iter_init(&iter, msg, len, 3);
+		g_isi_sb_iter_is_valid(&iter);
+		g_isi_sb_iter_next(&iter)) {
 
-	while (g_isi_sb_iter_is_valid(&iter)) {
+		struct ofono_network_operator *op;
+		char *tag = NULL;
+		guint8 taglen = 0;
+		guint8 status = 0;
+		guint8 umts = 0;
 
 		switch (g_isi_sb_iter_get_id(&iter)) {
 
-		case NET_AVAIL_NETWORK_INFO_COMMON: {
-			struct ofono_network_operator *op;
-			char *tag = NULL;
-			guint8 taglen = 0;
-			guint8 status = 0;
+		case NET_AVAIL_NETWORK_INFO_COMMON:
 
-			if (!g_isi_sb_iter_get_byte(&iter, &status, 2))
-				goto error;
-
-			if (!g_isi_sb_iter_get_byte(&iter, &taglen, 5))
-				goto error;
-
-			if (!g_isi_sb_iter_get_alpha_tag(&iter, &tag,
+			if (!g_isi_sb_iter_get_byte(&iter, &status, 2)
+				|| !g_isi_sb_iter_get_byte(&iter, &taglen, 5)
+				|| !g_isi_sb_iter_get_alpha_tag(&iter, &tag,
 						taglen * 2, 6))
 				goto error;
 
@@ -431,17 +440,17 @@ static bool available_resp_cb(GIsiClient *client, const void *restrict data,
 			op->name[OFONO_MAX_OPERATOR_NAME_LENGTH] = '\0';
 			g_free(tag);
 			break;
-		}
-
-		case NET_DETAILED_NETWORK_INFO: {
-			struct ofono_network_operator *op;
 
+		case NET_DETAILED_NETWORK_INFO:
 			op = list + detail++;
+
 			if (!g_isi_sb_iter_get_oper_code(&iter, op->mcc,
-								op->mnc, 2))
+								op->mnc, 2)
+				|| !g_isi_sb_iter_get_byte(&iter, &umts, 7))
 				goto error;
+
+			op->tech = umts ? 2 : 3;
 			break;
-		}
 
 		default:
 			DBG("Skipping sub-block: %s (%zu bytes)",
@@ -449,7 +458,6 @@ static bool available_resp_cb(GIsiClient *client, const void *restrict data,
 				g_isi_sb_iter_get_len(&iter));
 			break;
 		}
-		g_isi_sb_iter_next(&iter);
 	}
 
 	if (common == detail && detail == total) {
@@ -462,7 +470,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_list_operators(struct ofono_netreg *netreg,
@@ -495,8 +503,9 @@ error:
 	g_free(cbd);
 }
 
-static bool set_auto_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean set_auto_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -509,7 +518,7 @@ static bool set_auto_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (!msg || len < 3 || msg[0] != NET_SET_RESP)
-		goto error;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -525,7 +534,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_register_auto(struct ofono_netreg *netreg,
@@ -560,8 +569,9 @@ error:
 	g_free(cbd);
 }
 
-static bool set_manual_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean set_manual_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -569,13 +579,13 @@ static bool set_manual_resp_cb(GIsiClient *client, const void *restrict data,
 	struct netreg_data *nd = ofono_netreg_get_data(netreg);
 	ofono_netreg_register_cb_t cb = cbd->cb;
 
-	if(!msg) {
+	if (!msg) {
 		DBG("ISI client error: %d", g_isi_client_error(client));
 		goto error;
 	}
 
 	if (len < 3 || msg[0] != NET_SET_RESP)
-		goto error;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -591,7 +601,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_register_manual(struct ofono_netreg *netreg,
@@ -640,8 +650,9 @@ static void isi_deregister(struct ofono_netreg *netreg,
 	CALLBACK_WITH_FAILURE(cb, data);
 }
 
-static void rat_ind_cb(GIsiClient *client, const void *restrict data,
-			size_t len, uint16_t object, void *opaque)
+static void rat_ind_cb(GIsiClient *client,
+			const void *restrict data, size_t len,
+			uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_netreg *netreg = opaque;
@@ -684,8 +695,9 @@ static void rat_ind_cb(GIsiClient *client, const void *restrict data,
 	}
 }
 
-static bool rat_resp_cb(GIsiClient *client, const void *restrict data,
-			size_t len, uint16_t object, void *opaque)
+static gboolean rat_resp_cb(GIsiClient *client,
+				const void *restrict data, size_t len,
+				uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_netreg *netreg = opaque;
@@ -695,15 +707,15 @@ static bool rat_resp_cb(GIsiClient *client, const void *restrict data,
 
 	if (!msg) {
 		DBG("ISI client error: %d", g_isi_client_error(client));
-		return true;
+		return TRUE;
 	}
 
 	if (len < 3 || msg[0] != NET_RAT_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
-		return true;
+		return TRUE;
 	}
 
 	g_isi_sb_iter_init(&iter, msg, len, 3);
@@ -720,7 +732,7 @@ static bool rat_resp_cb(GIsiClient *client, const void *restrict data,
 				|| !info
 				|| !g_isi_sb_iter_get_byte(&iter,
 							&nd->gsm_compact, 4))
-				return true;
+				return TRUE;
 
 			break;
 		}
@@ -733,23 +745,66 @@ static bool rat_resp_cb(GIsiClient *client, const void *restrict data,
 		}
 		g_isi_sb_iter_next(&iter);
 	}
-	return true;
+	return TRUE;
 }
 
-static void rssi_ind_cb(GIsiClient *client, const void *restrict data,
-			size_t len, uint16_t object, void *opaque)
+static void rssi_ind_cb(GIsiClient *client,
+			const void *restrict data, size_t len,
+			uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_netreg *netreg = opaque;
+	struct netreg_data *nd = ofono_netreg_get_data(netreg);
 
 	if (!msg || len < 3 || msg[0] != NET_RSSI_IND)
 		return;
 
+	nd->strength = msg[1];
 	ofono_netreg_strength_notify(netreg, msg[1]);
 }
 
-static bool rssi_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static void time_ind_cb(GIsiClient *client,
+			const void *restrict data, size_t len,
+			uint16_t object, void *opaque)
+{
+	const unsigned char *msg = data;
+	const unsigned char *nitz = msg + 3;
+	struct ofono_netreg *netreg = opaque;
+
+	struct ofono_network_time info;
+
+	if (!msg || len < 13 || msg[0] != NET_TIME_IND
+		|| nitz[0] != NET_TIME_INFO)
+		return;
+
+	nitz += 2;
+
+	/* Value is years since last turn of century */
+	info.year = nitz[0] != NET_INVALID_TIME ? nitz[0] : -1;
+	info.year += 2000;
+
+	info.mon = nitz[1] != NET_INVALID_TIME ? nitz[1] : -1;
+	info.mday = nitz[2] != NET_INVALID_TIME ? nitz[2] : -1;
+	info.hour = nitz[3] != NET_INVALID_TIME ? nitz[3] : -1;
+	info.min = nitz[4] != NET_INVALID_TIME ? nitz[4] : -1;
+	info.sec = nitz[5] != NET_INVALID_TIME ? nitz[5] : -1;
+
+	/* Most significant bit set indicates negative offset. The
+	 * second most significant bit is 'reserved'. The value is the
+	 * offset from UTCin a count of 15min intervals, possibly
+	 * including the current DST adjustment. */
+	info.utcoff = (nitz[6] & 0x3F) * 15 * 60;
+	if (nitz[6] & 0x80)
+		info.utcoff *= -1;
+
+	info.dst = nitz[7] != NET_INVALID_TIME ? nitz[7] : -1;
+
+	ofono_netreg_time_notify(netreg, &info);
+}
+
+static gboolean rssi_resp_cb(GIsiClient *client,
+				const void *restrict data, size_t len,
+				uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -764,7 +819,7 @@ static bool rssi_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (len < 3 || msg[0] != NET_RSSI_GET_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s (0x%02X)",
@@ -805,7 +860,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_strength(struct ofono_netreg *netreg,
@@ -853,6 +908,7 @@ static gboolean isi_netreg_register(gpointer user)
 	g_isi_subscribe(nd->client, NET_REG_STATUS_IND, reg_status_ind_cb,
 			netreg);
 	g_isi_subscribe(nd->client, NET_RAT_IND, rat_ind_cb, netreg);
+	g_isi_subscribe(nd->client, NET_TIME_IND, time_ind_cb, netreg);
 
 	/* Bootstrap current RAT setting */
 	if (!g_isi_request_make(nd->client, rat, sizeof(rat),
@@ -865,13 +921,13 @@ static gboolean isi_netreg_register(gpointer user)
 	return FALSE;
 }
 
-static void reachable_cb(GIsiClient *client, bool alive, uint16_t object,
+static void reachable_cb(GIsiClient *client, gboolean alive, uint16_t object,
 				void *opaque)
 {
 	struct ofono_netreg *netreg = opaque;
 
 	if (!alive) {
-		DBG("Unable to bootsrap netreg driver");
+		DBG("Unable to bootstrap netreg driver");
 		return;
 	}
 
@@ -908,10 +964,12 @@ static void isi_netreg_remove(struct ofono_netreg *net)
 {
 	struct netreg_data *data = ofono_netreg_get_data(net);
 
-	if (data) {
-		g_isi_client_destroy(data->client);
-		g_free(data);
-	}
+	if (!data)
+		return;
+
+	ofono_netreg_set_data(net, NULL);
+	g_isi_client_destroy(data->client);
+	g_free(data);
 }
 
 static struct ofono_netreg_driver driver = {
